/**
 * Copyright (C) 2005-2009  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

import java.io.IOException;
import jCoCo.*;
import jAVR.*;

class T_CoCoUser
{
   // Konfigurationsdateien
   private SimpleXMLTransfer conf;
   private SimpleXMLTransfer hw;
   // Ausgabedateien
   private T_of_XML          konst;
   private T_of_raw          cmdreport;
   private T_of_raw          devices;
   private T_of_raw          codes;
   private jAVR              avr;
   
   // -------------------------------------------------------------------------
   
   public void init() throws XMLException, IOException
     {
       // --- Ausgabedateien erzeugen -----------------------------------------
       konst     = new T_of_XML(      "const.xml", null);
       cmdreport = new T_of_raw(      "cmd.html");
       devices   = new T_of_raw(      "devices.avr");
       codes     = new T_of_raw(      "codes.h");
       konst.forceWriteOnClose();
      
       // --- Eingabedateien öffnen -------------------------------------------
       conf   = new SimpleXMLTransfer("./../../src/autoconst/common.xml");
       hw     = new SimpleXMLTransfer("./../../src/autoconst/hw.xml");       
     }
   
   public void writeTargetBaudrate(double flfCPU,
                                   int    rate) throws IOException, XMLException
   {
     // -8 ist korrekt für UART_SendDelay
     // -8 ist korrekt für UART_RecDelay
     konst.write(new T_IntConst("autoconst.timing.targetUART.b" + rate,
                                (flfCPU/rate -8)/3.0, 1, 255));
   }
   
   public void writeCodes(SimpleXMLTransfer group,
                          String            name) throws IOException, XMLException
   {
     SimpleXMLTransfer item;
     String            itsname;
     
     int size = group.getChildCount();
     for (int n=0; n<size; n++)
     {
       item    = group.getChildAt(n);
       itsname = name+"_"+item.name();
       codes.write("#define " + itsname + " " + item.attribute("cmdbyte"));
       writeCodes(item, itsname);
     }
   }
      
   public void writeCmdReport() throws IOException, XMLException
   {
     cmdreport.write("<html>");
     cmdreport.write("<table cellspacing=1 cellpadding=1 border=1>");
     
     SimpleXMLTransfer group = conf.getChild("avrdev.host.cmd");
     SimpleXMLTransfer item;
     
     int size = group.getChildCount();
     for (int n=0; n<size; n++)
     {
       item = group.getChildAt(n);

       cmdreport.write("<tr>");
       
       cmdreport.write("<td>");
       cmdreport.write(item.name());
       cmdreport.write("</td>");

       cmdreport.write("<td>");
       cmdreport.write(item.attribute("cmdbyte"));
       cmdreport.write("</td>");

       cmdreport.write("<td>");
       cmdreport.write(item.attribute("status", "-"));
       cmdreport.write("</td>");
       
       cmdreport.write("<td>");
       cmdreport.write(item.getContentString());
       cmdreport.write("</td>");
       
       cmdreport.write("</tr>");
     }
     
     cmdreport.write("</table>");
     cmdreport.write("</html>");
   }
  
  public int GetDeviceFlags(SimpleXMLTransfer item) throws IOException, XMLException
  {
    int nPoll = 0;
    
    switch (item.getInt("pollmethod"))
    {
      case 1:
        nPoll = 1;
        break;
        
      case 0:
        nPoll = 0;
        break;
        
      default:
        T_ErrLog.add("pollmethod unbekannt");
        break;
    }    
    
    return(nPoll);
  }
  
   public void writeDevices() throws IOException, XMLException
   {
     int               size;
     SimpleXMLTransfer item;
     SimpleXMLTransfer group;
     int               sigbytes;
     
     group = conf.getChild("avrdev.devices_byte");
     devices.write("Table_Devices_byte:");     
     size = group.getChildCount();
     for (int n=0; n<size; n++)
     {
       item = group.getChildAt(n);
       sigbytes = item.getInt("sigbytes");
       devices.write(".db " + item.attributeAsInt("avr910devCode") + ", " 
                     + item.attribute("pollcode") + ", "
                     + GetDeviceFlags(item) + ", " 
                     + ((sigbytes>>16)&0xFF) + ", " + ((sigbytes>>8)&0xFF) + ", " + (sigbytes&0xFF)
                     + "; " + item.getContentString());
     }
     devices.write(".db 0");
     
     group = conf.getChild("avrdev.devices_page");
     devices.write("Table_Devices_page:");     
     size = group.getChildCount();
     for (int n=0; n<size; n++)
     {
       item = group.getChildAt(n);
       sigbytes = item.getInt("sigbytes");
       devices.write(".db " + item.attributeAsInt("avr910devCode") + ", "
                     + item.attribute("pageSizeWords") + ", "
                     + GetDeviceFlags(item) + ", "
                     + ((sigbytes>>16)&0xFF) + ", " + ((sigbytes>>8)&0xFF) + ", " + (sigbytes&0xFF)
                     + "; " + item.getContentString());
     }
     devices.write(".db 0");
   }
   
   public void calc() throws IOException, XMLException
     {
       switch (hw.getInt("avrdev.CPU.type"))
       {
        case 8535:
         avr = new jAVR_mega8535();
         break;
        case 16000000:
         avr = new jAVR_mega16();
         break;
        default:
         T_ErrLog.add("CPU-Typ bisher nicht eingepflegt.");
         break;
       }
                              
       writeCmdReport();
       writeDevices();       
       writeCodes(conf.getChild("avrdev.host.cmd"),   "CMD");
       writeCodes(conf.getChild("avrdev.host.reply"), "REPLY");
       
       
       double flfCPU    = hw.getDouble("avrdev.CPU.Frequenz");
       

       // Configurationsbits für Timer berechnen
       int prescalerVal = conf.getInt("avrdev.timer.timer0.prescalerVal");
       konst.write(new T_IntConst("autoconst.timer.timer0.prescaleSelect",
                                  avr.getTimer0PrescaleSelect(prescalerVal)));
       double d_TAwake = prescalerVal*256/flfCPU;
       System.out.println("awake every " + d_TAwake + " seconds");
       
       //
       konst.write(new T_IntConst("autoconst.timing.LEDCntInit",
                                  0.5/d_TAwake, 2, 255));

       // delay counts
       konst.write(new T_IntConst("autoconst.timing.T0_1ms",
                                  1.0E-3*256/d_TAwake, 2, 255));
       konst.write(new T_IntConst("autoconst.timing.T0_4ms",
                                  4.0E-3*256/d_TAwake, 2, 255));
       konst.write(new T_IntConst("autoconst.timing.T0_10ms",
                                  10.0E-3*256/d_TAwake, 2, 255));
       //
       prescalerVal = conf.getInt("avrdev.timer.timer2.prescalerVal");
       konst.write(new T_IntConst("autoconst.timer.timer2.prescaleSelect",
                                  avr.getTimer2PrescaleSelect(prescalerVal)));
       // cnt*prescalerVal/flfCPU = zeit
       // cnt = zeit*flfCPU/prescaleVal
       konst.write(new T_IntConst("autoconst.timing.timer2.TUARTRec",
                                  32.0E-3*flfCPU/prescalerVal, 1, 255));
       
       //
       writeTargetBaudrate(flfCPU,  9600);
       writeTargetBaudrate(flfCPU, 19200);

       // I want to have a clock of 0.5 MHz or less.
       // The mega128 datasheet says to use f_CPU/4 at most, 
       // f_CPU/6 at >=12MHz. This refers to the target CPU of course!
       // Using 0.5 MHz as f_SCLK limits the target CPUs clock to >= 2 MHz.
       double fl_f_SCLK = 100E3; // a safe default
       konst.write(new T_IntConst("autoconst.timing.SPIDelayCycles3",
                                  0.5+( flfCPU /2.0 /fl_f_SCLK /3.0), 0, 255));
       // Nun bekommt man übergeben: die Bitzeit in 100 Nanosekunden.
       // Es gilt     fl_f_SCLK = 1/(TBit100ns * 100E-9) = 10E6 / TBit100ns
       // Damit wird  flfCPU /2.0 /fl_f_SCLK /3.0
       // zu          flfCPU /2.0 *TBit100ns /10E6 /3.0
       // Das muss man jetzt im uC rechnen. Ich versuche mal zu vereinfachen (nicht
       // exakt, aber OK).
       // TBit100ns * flfCPU/60E6
       int schieb = 0;
       while ((1<<schieb)*flfCPU/60E6 < 0.5)
         schieb++;
       konst.write(new T_IntConst("autoconst.timing.SPIDelayShift",
                                  schieb, 0, 255));
       
       // UART
       {
         long nVal = Math.round(flfCPU /(16*conf.getFloat("avrdev.UART.baudrate"))-1);
         konst.write(new T_IntConst("autoconst.UART.ubrr", nVal, 0, 255));           
       }
       
       // ADC, scaling
       // incorrect results when blocking this part!
       //{
         // nADC is the ADC result (8 significant bit)
         // nU is the value representing the voltage, 256 <=> U_max
         // 
         // 1. nADC = 256 * U * R2 / (R1+R2) / U_ref
         //       U = nADC * (R1+R2) * U_ref / 256 / R2
         // 2. nU   = 256 * U / U_max
         // 
         // ->
         // 
         // nU   = 256 * nADC * (R1+R2) * U_ref / 256 / R2 / U_max
         // nU   =       nADC * (R1+R2) * U_ref       / R2 / U_max
         double R1 = hw.getDouble("avrdev.adc.R1");
         double R2 = hw.getDouble("avrdev.adc.R2");
         
         konst.write(new T_MulShift("autoconst.msADC_U",
                                    (R1+R2) * hw.getDouble("avrdev.adc.U_ref")
                                    / R2
                                    / hw.getDouble("avrdev.adc.U_max"),
                                    8
                                    ));
         
         // input range
         System.out.println("voltage input range: " 
                            + (hw.getDouble("avrdev.adc.U_ref") * (R1+R2) / R2)
                            + " V");
       
         konst.write(new T_IntConst("autoconst.voltage.max", 
                                    256*conf.getDouble("avrdev.voltage.U_max")/hw.getDouble("avrdev.adc.U_max"),
                                    1, 255));           
         konst.write(new T_IntConst("autoconst.voltage.soll",
                                    256*conf.getDouble("avrdev.voltage.U_soll")/hw.getDouble("avrdev.adc.U_max"),
                                    1, 255));
       
         codes.write("#define REFVOLTAGE256  " + hw.getString("avrdev.adc.U_max"));
       
       // incorrect results when blocking this part!
       //}
                            
       // Pins
       SimpleXMLTransfer pins = hw.getChild("avrdev.pins");
       for (int i=0; i<pins.getChildCount(); i++)
         {
           char c = pins.getChildAt(i).getString("port").toUpperCase().charAt(0);
           konst.write(new T_PreProDef("autoconst.pin." + pins.getChildAt(i).name() + ".dir", 
                                       "DDR"  + c + ", "  + pins.getChildAt(i).getString("bit")));
           konst.write(new T_PreProDef("autoconst.pin." + pins.getChildAt(i).name() + ".data",
                                       "PORT" + c + ", " + pins.getChildAt(i).getString("bit")));
           konst.write(new T_PreProDef("autoconst.pin." + pins.getChildAt(i).name() + ".in",
                                       "PIN"  + c + ", " + pins.getChildAt(i).getString("bit")));
           konst.write(new T_PreProDef("autoconst.pin." + pins.getChildAt(i).name() + ".inport",
                                       "PIN"  + c));
           konst.write(new T_PreProDef("autoconst.pin." + pins.getChildAt(i).name() + ".dirport",
                                       "DDR"  + c));
           konst.write(new T_PreProDef("autoconst.pin." + pins.getChildAt(i).name() + ".dataport",
                                       "PORT"  + c));
         }
     }

   // -------------------------------------------------------------------------
   
   public void close() throws IOException
     {
       konst.close();
       cmdreport.close();
       devices.close();
       codes.close();
     }

   // -------------------------------------------------------------------------
   // -------------------------------------------------------------------------
         
   public static void main(String args[])
     {
      T_ErrLog.init();
      T_CoCoUser  user = new T_CoCoUser();

      try
        {
          user.init();
          user.calc();
          user.close();
        }
      catch (XMLException ce)
        {
          T_ErrLog.add("Error in configuration file: " + ce.getMessage());
        }
      catch (IOException ce)
        {
          T_ErrLog.add(ce.getMessage());
        }
      
      switch (T_ErrLog.getCnt())
        {
         case 0:
          System.out.println("No errors detected");
          break;
         case 1:
          System.out.println("1 error detected");
          System.exit(-1);
          break;
         default:
          System.out.println(T_ErrLog.getCnt() + " errors detected");
          System.exit(-1);
        }
     }
}
